-- Written by Team Ulysses, http://ulyssesmod.net/
AddCSLuaFile( "autorun/uclip.lua" )

-- Config
local function adminCheck( ply ) -- Who can move through everything? Comment out these lines for no one.
	if ulx and ply:query( "ulx noclip" ) then 
		return true
	elseif ply:IsAdmin() or ply:IsSuperAdmin() then 
		return true 
	end
	
	return false
end
-- End config

local hasPropProtection = false -- Chaussette's Prop Protection (preferred over PropSecure)
local propProtectionFn -- Function to call to see if a prop belongs to a player. We have to fetch it from a local so we'll store it here.

local hasPropSecure = false -- Prop Secure by Conna
local hasProtector = false -- Protector by Conna

local noProtection = false -- If there's no protection whatsoever, this is flagged. 
-- We need this flag because with a protector, we default to _not_ being able to go through things.
-- This flag saves us major memory/bandwidth when there's no protection

if SERVER then
	-- We'll check status of protectors in this init
	local function init()
		local t = hook.GetTable()
		local fn
		if t.CanTool[ 0 ] then -- ULib
			fn = t.CanTool[ 0 ].PropProtection
		else
			fn = t.CanTool.PropProtection
		end
		
		hasPropProtection = type( fn ) == "function"
		
		if hasPropProtection then
			-- We're going to get the function we need now. It's local so this is a bit dirty			
			local gi = debug.getinfo( fn )		
			for i=1,gi.nups do
			    if debug.getupvalue( fn, i ) == "Appartient" then
			    	local junk
			        junk, propProtectionFn = debug.getupvalue( fn, i )
			        break
			    end
			end
		end
		
		hasPropSecure = type( PropSecure ) == "table"
		hasProtector = type( Protector ) == "table"
		
		if not hasPropProtection and not hasPropSecure and not hasProtector then
			noProtection = true
			umsg.Start( "UclipNoProtection" ) -- In case there's anyone connected right now.
			umsg.End()
		end
	end
	hook.Add( "Initialize", "UclipInitialize", init )
	
	-- Tell them there's no prop protection in the even there isn't.
	local function initialSpawn( ply )
		if noProtection then
			umsg.Start( "UclipNoProtection", ply )
			umsg.End()
		end
	end
	hook.Add( "PlayerInitialSpawn", "UclipInitialSpawn", initialSpawn )
	
	-- Block the valve noclip-unstuck. Can go through walls.
	local function noclip( ply )
		if ply:GetMoveType() == MOVETYPE_NOCLIP then
			local pos = ply:GetPos()
			timer.Simple( FrameTime() * 0.5, function() 
				ply:SetPos( pos )
			end )
		end
	end
	hook.Add( "PlayerNoClip", "UclipNoclip", noclip )	
end

-- This function checks the protector to see if ownership has changed from what we think it is. Notifies player too.	
local function updateOwnership( ply, ent )
	if noProtection then return end -- No point on going on	
	if not ent.Uclip then ent.Uclip = {} end -- Initialize table
	if CLIENT then return end -- Server stuff from here on out

	local owns
	if hasPropProtection then -- Chaussette's Prop Protection (preferred over PropSecure)
		owns = propProtectionFn( ply, ent )
	elseif hasPropSecure then -- PropSecure
		owns = PropSecure.IsPlayers( ply, ent )
	elseif hasProtector then -- Protector
		owns = Protector.Owner( ent ) == ply:UniqueID()
	end
	
	if owns == false then -- More convienent to store as nil, takes less memory!
		owns = nil
	end
	
	if ent.Uclip[ ply ] ~= owns then
		ent.Uclip[ ply ] = owns
		umsg.Start( "UclipOwnershipUpdate", ply )
			umsg.Entity( ent )
			umsg.Bool( owns )
		umsg.End()
	end
end

if CLIENT then
	local function rcvOwnershipUpdate( um )
		local ent = um:ReadEntity()
		if not ent or not ent:IsValid() then return end
		
		updateOwnership( LocalPlayer(), ent ) -- So it sets up the table if we don't have it already
		local owns = um:ReadBool()
		if owns == false then -- More convienent to store as nil, takes less memory!
			owns = nil
		end		
		ent.Uclip[ LocalPlayer() ] = owns
	end
	usermessage.Hook( "UclipOwnershipUpdate", rcvOwnershipUpdate )
	
	local function rcvNoProtection( um )
		noProtection = true
	end
	usermessage.Hook( "UclipNoProtection", rcvNoProtection )	
end
	
local function checkOwnership( ply, ent )
	if noProtection then return true end -- No protection, they own everything.
	
	updateOwnership( ply, ent )	-- Make sure server and the client are current
	return ent.Uclip[ ply ]
end

function calcWallSlide( vel, normal )
	local toWall = normal:GetNormalized() * -1
	local dot = vel:GetNormalized():Dot( toWall )
	
	if dot > 0.999 then return Vector( 0, 0, 0 ) end -- No slide, too similar.
	if dot <= 0 then return Vector( 0, 0, 0 ) end -- Shouldn't ever get this but just in case
	
	-- Figure out vel to wall using cos( theta ) = adjacent/hypotenuse, then times it by direction
	local velToWall = dot * vel:Length() * toWall
	return vel - velToWall
end

local function checkVel( ply, move, vel )
	local ft = FrameTime()
	
	local t = {}
	t.start = ply:GetPos()
	t.endpos = vel * ft + (vel:GetNormalized() * 64)
	t.endpos = t.endpos + ply:GetPos()
	t.filter = { ply }
	local tr = util.TraceEntity( t, ply )

	while tr.Hit do -- Recursively check all the hits. This is so we don't miss objects behind another object.
		if (CLIENT or ply:IsInWorld()) and tr.HitWorld and tr.HitPos == tr.StartPos then -- You can get stuck if you crouch and then stand up in a small area and then you crash the client/server. 
			--Just leave em stuck and label this "todo". Seems like they can get out by un-noclipping anyways.
			move:SetVelocity( Vector( 0, 0, 0 ) )
			move:SetForwardSpeed( 0 )
			move:SetSideSpeed( 0 )
			return
		end 
		
		if tr.HitWorld or ( tr.Entity:IsValid() and tr.HitPos ~= tr.StartPos and not checkOwnership( ply, tr.Entity ) ) then -- If world or a prop they don't own that they're not stuck inside.
			local slide = calcWallSlide( vel, tr.HitNormal )
			move:SetVelocity( slide )			
			move:SetSideSpeed( 0 )
			move:SetForwardSpeed( 0 )
			
			return checkVel( ply, move, slide ) -- Return now so this func isn't left on stack
		end

		if tr.Entity and tr.Entity:IsValid() then -- Ent to add!
			table.insert( t.filter, tr.Entity )
		end

		tr = util.TraceEntity( t, ply )
	end
end

local function move( ply, move )
	if ply:GetMoveType() ~= MOVETYPE_NOCLIP then return end -- Not interested in anything but noclip
	if GetConVarNumber( "sbox_noclip" ) == 2 or ply:query( "ulx noclip" ) then return end -- They can go through everything

	local ft = FrameTime()

	local vel = move:GetVelocity() -- How far are they going to move this frame?
	vel = vel + ply:EyeAngles():Forward() * move:GetForwardSpeed() / 4 * ft
	vel = vel + ply:EyeAngles():Right() * move:GetSideSpeed() / 4 * ft
	-- We're not doing any extra checks we should be here (approaching max, slowing down, etc) because it really doesn't matter enough to us.
	
	checkVel( ply, move, vel )
end
hook.Add( "Move", "UclipMove", move )